import sys
import os
from rich import print as rprint


def exit_cli(message):
    rprint("%s" % message)
    sys.exit(0)

def display_file(file_name):
    """
    Open given file with default user program.
    """
    if sys.platform.startswith('linux'):
        os.system('xdg-open %s' % file_name)

    elif sys.platform.startswith('darwin'):
        os.system('open %s' % file_name)



from satsearch import Search
import sys
from rich import print as rprint

from felicette.utils.geo_utils import get_tiny_bbox
from felicette.utils.sys_utils import exit_cli
from felicette.constants import band_tag_map
from felicette.utils.file_manager import (
    save_to_file,
    data_file_exists,
    file_paths_wrt_id,
)


def handle_prompt_response(response):
    if response in ["n", "N"]:
        exit_cli(
            "Why not try a different location next time? I'd suggest [link=https://en.wikipedia.org/wiki/Svalbard]Svalbard[/link] :)"
        )
    elif response in ["y", "Y", ""]:
        return None
    else:
        exit_cli("[red]Sorry, invalid response. Exiting :([/red]")


def search_landsat_data(coordinates, cloud_cover_lt):
    search = Search(
        bbox=get_tiny_bbox(coordinates),
        query={
            "eo:cloud_cover": {"lt": cloud_cover_lt},
            "collection": {"eq": "landsat-8-l1"},
        },
        sort=[{"field": "eo:cloud_cover", "direction": "asc"}],
    )

    # improvement: filter by date, cloud cover here

    search_items = search.items()
    if not len(search_items):
        exit_cli("No data matched your search, please try different parameters.")
    landsat_item = search_items[0]
    return landsat_item


def preview_landsat_image(landsat_item):
    paths = file_paths_wrt_id(landsat_item._data["id"])
    # download image and save it in directory
    if not data_file_exists(paths["preview"]):
        save_to_file(
            landsat_item.assets["thumbnail"]["href"],
            paths["preview"],
            landsat_item._data["id"],
            "✗ preview data doesn't exist, downloading image",
        )
    else:
        rprint("[green] ✓ ", "required data exists for preview image")
    # print success info
    rprint("[blue]Preview image saved at:[/blue]")
    print(paths["preview"])
    # prompt a confirm option
    response = input(
        "Are you sure you want to see an enhanced version of the image at the path shown above? [Y/n]"
    )
    return handle_prompt_response(response)


def download_landsat_data(landsat_item, bands):

    # get paths w.r.t. id
    paths = file_paths_wrt_id(landsat_item._data["id"])
    # save bands generically
    for band in bands:
        band_filename = paths["b%s" % band]
        if not data_file_exists(band_filename):
            save_to_file(
                landsat_item.assets["B{}".format(band)]["href"],
                band_filename,
                landsat_item._data["id"],
                "✗ required data doesn't exist, downloading %s %s"
                % (band_tag_map["b" + str(band)], "band"),
            )
        else:
            rprint(
                "[green] ✓ ",
                "required data exists for {} band".format(
                    band_tag_map["b" + str(band)]
                ),
            )

    return landsat_item._data["id"]



import click
import sys

from felicette.utils.geo_utils import geocoder_util
from felicette.utils.file_manager import check_sat_path
from felicette.sat_downloader import (
    download_landsat_data,
    search_landsat_data,
    preview_landsat_image,
)
from felicette.utils.sys_utils import exit_cli
from felicette.sat_processor import process_landsat_data


@click.command()
@click.option(
    "-c",
    "--coordinates",
    nargs=2,
    type=float,
    help="Coordinates in (lon, lat) format. This overrides -l command",
)
@click.option("-l", "--location-name", type=str, help="Location name in string format")
@click.option(
    "-p",
    "--pan-enhancement",
    default=False,
    is_flag=True,
    help="Enhance image with panchromatic band",
)
@click.option(
    "-pre",
    "--preview-image",
    default=False,
    is_flag=True,
    help="Preview pre-processed low resolution RGB satellite image.",
)
@click.option(
    "-v",
    "--vegetation",
    default=False,
    is_flag=True,
    help="Show Color Infrared image to highlight vegetation",
)
def main(coordinates, location_name, pan_enhancement, preview_image, vegetation):
    """Satellite imagery for dummies."""
    if not coordinates and not location_name:
        exit_cli("Please specify either --coordinates or --location-name")
    if location_name:
        coordinates = geocoder_util(location_name)

    # unless specified, cloud_cover_lt is 10
    landsat_item = search_landsat_data(coordinates, 10)

    # check if directory exists to save the data for this product id
    check_sat_path(landsat_item._data["id"])

    # if preview option is set, download and preview image
    if preview_image:
        preview_landsat_image(landsat_item)

    # set bands to process
    bands = [2, 3, 4]
    if pan_enhancement:
        bands.append(8)

    if vegetation:
        bands = [3, 4, 5]

    # NB: can't enable pan-enhancement with vegetation

    # download data
    data_id = download_landsat_data(landsat_item, bands)
    # process data
    process_landsat_data(data_id, bands)


if __name__ == "__main__":
    main()



import requests


def find_first_landsat(items):
    for index, item in enumerate(items):
        if "LC" in item._data["id"]:
            return item


def geocoder_util(location_name):
    # call nominatim api
    r = requests.get(
        "https://nominatim.openstreetmap.org/search?city={}&format=json".format(
            location_name
        )
    )
    r_json = r.json()
    # return lat, lon
    return (float(r_json[0]["lon"]), float(r_json[0]["lat"]))


def get_tiny_bbox(coordinates):
    lon, lat = coordinates
    return [lon, lat, lon + 0.00001, lat + 0.00001]



"""
Source: https://github.com/mapbox/rio-color/blob/master/rio_color/scripts/cli.py

Further modifications made to make it functional with felicette
"""
import rasterio
from rasterio.rio.options import creation_options
from rasterio.transform import guard_transform
from rio_color.workers import atmos_worker, color_worker
from rio_color.operations import parse_operations, simple_atmo_opstring
import riomucho
from rich import print


def check_jobs(jobs):
    """Validate number of jobs."""
    if jobs == 0:
        print("Jobs must be >= 1 or == -1")
    elif jobs < 0:
        import multiprocessing

        jobs = multiprocessing.cpu_count()
    return jobs


def color(jobs, out_dtype, src_path, dst_path, operations, creation_options):
    with rasterio.open(src_path) as src:
        opts = src.profile.copy()
        windows = [(window, ij) for ij, window in src.block_windows()]

    opts.update(**creation_options)
    opts["transform"] = guard_transform(opts["transform"])

    out_dtype = out_dtype if out_dtype else opts["dtype"]
    opts["dtype"] = out_dtype

    args = {"ops_string": " ".join(operations), "out_dtype": out_dtype}
    # Just run this for validation this time
    # parsing will be run again within the worker
    # where its returned value will be used
    try:
        parse_operations(args["ops_string"])
    except ValueError as e:
        import sys

        sys.exit(1)
        print(e)

    jobs = check_jobs(jobs)

    if jobs > 1:
        with riomucho.RioMucho(
            [src_path],
            dst_path,
            color_worker,
            windows=windows,
            options=opts,
            global_args=args,
            mode="manual_read",
        ) as mucho:
            mucho.run(jobs)
    else:
        with rasterio.open(dst_path, "w", **opts) as dest:
            with rasterio.open(src_path) as src:
                rasters = [src]
                for window, ij in windows:
                    arr = color_worker(rasters, window, ij, args)
                    dest.write(arr, window=window)

                dest.colorinterp = src.colorinterp



import rasterio as rio
import numpy as np
from rio_color import operations, utils
from PIL import Image
import PIL
from rich import print as rprint

from felicette.utils.color import color
from felicette.utils.gdal_pansharpen import gdal_pansharpen
from felicette.utils.file_manager import file_paths_wrt_id
from felicette.utils.image_processing_utils import process_sat_image
from felicette.utils.sys_utils import display_file

# increase PIL image processing pixels count limit
PIL.Image.MAX_IMAGE_PIXELS = 933120000


def process_landsat_vegetation(id, bands):

    # get paths of files related to this id
    paths = file_paths_wrt_id(id)

    # stack NIR, R, G bands

    # open files from the paths, and save it as stack
    b5 = rio.open(paths["b5"])
    b4 = rio.open(paths["b4"])
    b3 = rio.open(paths["b3"])

    # read as numpy ndarrays
    nir = b5.read(1)
    r = b4.read(1)
    g = b3.read(1)

    with rio.open(
        paths["stack"],
        "w",
        driver="Gtiff",
        width=b4.width,
        height=b4.height,
        count=3,
        crs=b4.crs,
        transform=b4.transform,
        dtype=b4.dtypes[0],
        photometric="RGB",
    ) as rgb:
        rgb.write(nir, 1)
        rgb.write(r, 2)
        rgb.write(g, 3)
        rgb.close()

    source_path_for_rio_color = paths["stack"]

    rprint("Let's make our 🌍 imagery a bit more colorful for a human eye!")
    # apply rio-color correction
    ops_string = "sigmoidal rgb 20 0.2"
    # refer to felicette.utils.color.py to see the parameters of this function
    # Bug: number of jobs if greater than 1, fails the job
    color(
        1,
        "uint16",
        source_path_for_rio_color,
        paths["vegetation_path"],
        ops_string.split(","),
        {"photometric": "RGB"},
    )

    # resize and save as jpeg image
    print("Generated 🌍 images!🎉")
    rprint("[yellow]Please wait while I resize and crop the image :) [/yellow]")
    process_sat_image(paths["vegetation_path"], paths["vegetation_path_jpeg"])
    rprint("[blue]GeoTIFF saved at:[/blue]")
    print(paths["vegetation_path"])
    rprint("[blue]JPEG image saved at:[/blue]")
    print(paths["vegetation_path_jpeg"])
    # display generated image
    display_file(paths["vegetation_path_jpeg"])


def process_landsat_rgb(id, bands):
    # get paths of files related to this id
    paths = file_paths_wrt_id(id)

    # stack R,G,B bands

    # open files from the paths, and save it as stack
    b4 = rio.open(paths["b4"])
    b3 = rio.open(paths["b3"])
    b2 = rio.open(paths["b2"])

    # read as numpy ndarrays
    r = b4.read(1)
    g = b3.read(1)
    b = b2.read(1)

    with rio.open(
        paths["stack"],
        "w",
        driver="Gtiff",
        width=b4.width,
        height=b4.height,
        count=3,
        crs=b4.crs,
        transform=b4.transform,
        dtype=b4.dtypes[0],
        photometric="RGB",
    ) as rgb:
        rgb.write(r, 1)
        rgb.write(g, 2)
        rgb.write(b, 3)
        rgb.close()

    source_path_for_rio_color = paths["stack"]

    # check if band 8, i.e panchromatic band has to be processed
    if 8 in bands:
        # pansharpen the image
        rprint(
            "Pansharpening image, get ready for some serious resolution enhancement! ✨"
        )
        gdal_pansharpen(["", paths["b8"], paths["stack"], paths["pan_sharpened"]])
        # set color operation's path to the pansharpened-image's path
        source_path_for_rio_color = paths["pan_sharpened"]

    rprint("Let's make our 🌍 imagery a bit more colorful for a human eye!")
    # apply rio-color correction
    ops_string = "sigmoidal rgb 20 0.2"
    # refer to felicette.utils.color.py to see the parameters of this function
    # Bug: number of jobs if greater than 1, fails the job
    color(
        1,
        "uint16",
        source_path_for_rio_color,
        paths["output_path"],
        ops_string.split(","),
        {"photometric": "RGB"},
    )

    # resize and save as jpeg image
    print("Generated 🌍 images!🎉")
    rprint("[yellow]Please wait while I resize and crop the image :) [/yellow]")
    process_sat_image(paths["output_path"], paths["output_path_jpeg"])
    rprint("[blue]GeoTIFF saved at:[/blue]")
    print(paths["output_path"])
    rprint("[blue]JPEG image saved at:[/blue]")
    print(paths["output_path_jpeg"])
    # display generated image
    display_file(paths["output_path_jpeg"])

def process_landsat_data(id, bands):

    if bands == [2, 3, 4] or bands == [2, 3, 4, 8]:
        process_landsat_rgb(id, bands)
    elif bands == [3, 4, 5]:
        process_landsat_vegetation(id, bands)



import tempfile
import os
import requests
from tqdm import tqdm
from rich import print as rprint

from felicette.constants import band_tag_map

workdir = os.path.join(os.path.expanduser("~"), "felicette-data")


def check_sat_path(id):
    data_path = os.path.join(workdir, id)

    if not os.path.exists(data_path):
        os.makedirs(data_path, exist_ok=True)


def save_to_file(url, filename, id, info_message):
    data_path = os.path.join(workdir, id)
    data_id = filename.split("/")[-1].split("-")[1].split(".")[0]
    rprint(info_message)
    file_path = os.path.join(data_path, filename)
    response = requests.get(url, stream=True)
    with tqdm.wrapattr(
        open(file_path, "wb"),
        "write",
        miniters=1,
        desc=data_id,
        total=int(response.headers.get("content-length", 0)),
    ) as fout:
        for chunk in response.iter_content(chunk_size=4096):
            fout.write(chunk)
    fout.close()


def data_file_exists(filename):
    return os.path.exists(filename)


def file_paths_wrt_id(id):
    home_path_id = os.path.join(workdir, id)
    return {
        "base": home_path_id,
        "preview": os.path.join(home_path_id, "%s-preview.jpg" % (id)),
        "b5": os.path.join(home_path_id, "%s-b5.tiff" % (id)),
        "b4": os.path.join(home_path_id, "%s-b4.tiff" % (id)),
        "b3": os.path.join(home_path_id, "%s-b3.tiff" % (id)),
        "b2": os.path.join(home_path_id, "%s-b2.tiff" % (id)),
        "b8": os.path.join(home_path_id, "%s-b8.tiff" % (id)),
        "stack": os.path.join(home_path_id, "%s-stack.tiff" % (id)),
        "pan_sharpened": os.path.join(home_path_id, "%s-pan.tiff" % (id)),
        "output_path": os.path.join(home_path_id, "%s-color-processed.tiff" % (id)),
        "output_path_jpeg": os.path.join(
            home_path_id, "%s-color-processed.jpeg" % (id)
        ),
        "vegetation_path": os.path.join(home_path_id, "%s-vegetation.tiff" % (id)),
        "vegetation_path_jpeg": os.path.join(home_path_id, "%s-vegetation.jpeg" % (id)),
    }



import pytest
from click.testing import CliRunner
from felicette import cli


@pytest.fixture
def runner():
    return CliRunner()


def test_cli(runner):
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert not result.exception
    assert result.output.strip() == "Hello, world."


def test_cli_with_option(runner):
    result = runner.invoke(cli.main, ["--as-cowboy"])
    assert not result.exception
    assert result.exit_code == 0
    assert result.output.strip() == "Howdy, world."


def test_cli_with_arg(runner):
    result = runner.invoke(cli.main, ["Shivashis"])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output.strip() == "Hello, Shivashis."



"""
Satellite imagery for dummies.
"""
from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

dependencies = [
    "click==7.1.2",
    "requests==2.24.0",
    "sat-search==0.2.3",
    "rich==3.3.2",
    "tqdm==4.48.0",
    "rasterio==1.1.5",
    "rio-color==1.0.0",
    "pillow==7.2.0",
    "opencv-python==4.3.0.36"
]

setup_dependencies = ['numpy']

setup(
    name="felicette",
    version="0.1.7",
    url="https://github.com/plant99/felicette",
    license="BSD",
    author="Shivashis Padhi",
    author_email="shivashispadhi@gmail.com",
    description="Satellite imagery for dummies.",
    long_description=__doc__,
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    setup_requires=setup_dependencies,
    install_requires=dependencies,
    entry_points={"console_scripts": ["felicette = felicette.cli:main",],},
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # "Development Status :: 2 - Pre-Alpha",
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# import the necessary packages
import numpy as np
import cv2


def find_max_area_index(contours):
    contour_areas = list(map(cv2.contourArea, contours))
    return contour_areas.index(max(contour_areas))


def straighten_image(image):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    gray = cv2.bitwise_not(gray)
    # threshold the image, setting all foreground pixels to 255 and all background pixels to 0
    thresh = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)[1]
    # flip data in threshold array
    thresh1 = thresh - 255
    np.where(thresh1 == -255, 0, thresh1)

    # find angle of inclination

    coords = np.column_stack(np.where(thresh1 > 0))
    angle = cv2.minAreaRect(coords)[-1]
    if angle < -45:
        angle = -(90 + angle)
    else:
        angle = -angle

    # rotate the image to deskew it
    (h, w) = image.shape[:2]
    center = (w // 2, h // 2)
    M = cv2.getRotationMatrix2D(center, angle, 1.0)
    rotated = cv2.warpAffine(
        image, M, (w, h), flags=cv2.INTER_CUBIC, borderMode=cv2.BORDER_REPLICATE
    )
    return rotated


def remove_margin(rotated):
    # crop image to remove black margin
    gray = cv2.cvtColor(rotated, cv2.COLOR_BGR2GRAY)
    _, thresh = cv2.threshold(gray, 1, 255, cv2.THRESH_BINARY)
    # # Now find contours in it. There will be only one object, so find bounding rectangle for it.

    contours, hierarchy = cv2.findContours(
        thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    max_contour_index = find_max_area_index(contours)
    cnt = contours[max_contour_index]
    x, y, w, h = cv2.boundingRect(cnt)
    # # Now crop the image, and save it into another file.

    crop = rotated[y : y + h, x : x + w]
    return crop


def process_sat_image(source_path, dest_path):
    # load the image from disk
    image = cv2.imread(source_path)
    straightened_image = straighten_image(image)
    cropped_image = remove_margin(straightened_image)

    # write jpeg to destination
    cv2.imwrite(dest_path, cropped_image)
